CREATE OR REPLACE FUNCTION __API_SCHEMA_INTERNAL_V2__.__EXTENSION_OBJECT_V2__(_core_bson_to_bson)(__CORE_SCHEMA_V2__.bson)
 RETURNS __CORE_SCHEMA__.bson
 LANGUAGE c
 IMMUTABLE PARALLEL SAFE STRICT
AS 'MODULE_PATHNAME', __CONCAT_NAME_FUNCTION__($function$, __EXTENSION_OBJECT_PREFIX_V2__, _core_bson_to_bson$function$);