/*-------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * include/documentdb_distributed_init.h
 *
 * The implementation for shard colocation for documentdb with citus.
 *
 *-------------------------------------------------------------------------
 */
#ifndef DOCUMENTDB_DISTRIBUTED_INIT_H
#define DOCUMENTDB_DISTRIBUTED_INIT_H

void InitDocumentDBDistributedConfigurations(const char *prefix);

#endif
