
ALTER TYPE __API_CATALOG_SCHEMA__.index_spec_type_internal ADD ATTRIBUTE cosmos_search_options __CORE_SCHEMA__.bson;
ALTER TYPE __API_CATALOG_SCHEMA__.index_spec_type_internal ADD ATTRIBUTE index_options __CORE_SCHEMA__.bson;