/*-------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * include/shard_colocation.h
 *
 * The implementation for shard colocation for DocumentDB with citus.
 *
 *-------------------------------------------------------------------------
 */
#ifndef SHARD_COLOCATION_H
#define SHARD_COLOCATION_H

void UpdateColocationHooks(void);

#endif
