/*-------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * include/distributed_hooks.h
 *
 * The implementation for shard colocation for DocumentDB with citus.
 *
 *-------------------------------------------------------------------------
 */
#ifndef DOCUMENTDB_DISTRIBUTED_HOOKS_H
#define DOCUMENTDB_DISTRIBUTED_HOOKS_H

void InitializeDocumentDBDistributedHooks(void);

#endif
