
/*
 * __API_SCHEMA_V2__.find_and_modify processes a mongo wire protocol findAndModify command.
 */
CREATE OR REPLACE FUNCTION __API_SCHEMA_V2__.find_and_modify(
    p_database_name text,
    p_message __CORE_SCHEMA_V2__.bson,
    p_transaction_id text default NULL,
    p_result OUT __CORE_SCHEMA_V2__.bson,
    p_success OUT boolean)
 RETURNS record
 LANGUAGE C
 VOLATILE
AS 'MODULE_PATHNAME', $$command_find_and_modify$$;
