-- show all functions exported in documentdb_api_distributed
\df documentdb_api_distributed.*

-- show all aggregates exported
\da+ documentdb_api_distributed.*

-- This is last (Tables/indexes)
\d documentdb_api_distributed.*
