/*-------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * include/commands/cursor_common.h
 *
 * Common declarations of Cursor commands.
 *
 *-------------------------------------------------------------------------
 */

extern const StringView CursorContinuationTableName;
extern const StringView CursorContinuationValue;
extern const StringView PrimaryKeyShardKey;
