CREATE OPERATOR CLASS __API_CATALOG_SCHEMA__.bson_rum_wildcard_project_path_ops
    FOR TYPE __CORE_SCHEMA__.bson using __EXTENSION_OBJECT__(_rum) AS
        OPERATOR        1       __API_CATALOG_SCHEMA__.@= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        2       __API_CATALOG_SCHEMA__.@> (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        3       __API_CATALOG_SCHEMA__.@>= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        4       __API_CATALOG_SCHEMA__.@< (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        5       __API_CATALOG_SCHEMA__.@<= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        6       __API_CATALOG_SCHEMA__.@*= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        7       __API_CATALOG_SCHEMA__.@!= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        8       __API_CATALOG_SCHEMA__.@!*= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR        9       __API_CATALOG_SCHEMA__.@~ (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       10       __API_CATALOG_SCHEMA__.@? (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       11       __API_CATALOG_SCHEMA__.@@# (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       12       __API_CATALOG_SCHEMA__.@# (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       13       __API_CATALOG_SCHEMA__.@&= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
--      skip unique index equal for wildcard since unique is not supported for wildcard
--      OPERATOR       14       __API_CATALOG_SCHEMA__.=?= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       15       __API_CATALOG_SCHEMA__.@!& (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       16       __API_CATALOG_SCHEMA__.@!| (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       17       __API_CATALOG_SCHEMA__.@#? (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       18       __API_CATALOG_SCHEMA__.@& (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       19       __API_CATALOG_SCHEMA__.@| (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        OPERATOR       20       __API_CATALOG_SCHEMA__.@% (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson),
        FUNCTION        1       __API_CATALOG_SCHEMA__.gin_bson_compare(bytea, bytea),
        FUNCTION        2       __API_CATALOG_SCHEMA__.gin_bson_wildcard_project_extract_value(__CORE_SCHEMA__.bson, internal),
        FUNCTION        3       __API_CATALOG_SCHEMA__.gin_bson_extract_query(__CORE_SCHEMA__.bson, internal, int2, internal, internal, internal, internal),
        FUNCTION        4       __API_CATALOG_SCHEMA__.gin_bson_consistent(internal, int2, anyelement, int4, internal, internal),
        FUNCTION        5       __API_CATALOG_SCHEMA__.gin_bson_compare_partial(bytea, bytea, int2, internal),
        FUNCTION        11       (__CORE_SCHEMA__.bson) __API_CATALOG_SCHEMA__.gin_bson_wildcard_project_options(internal),
    STORAGE         bytea;