ALTER OPERATOR __API_CATALOG_SCHEMA__.@=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@!=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@<(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@<=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@>(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@>=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@*=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@!*=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@~(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@%(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@?(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@@#(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@#(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@&=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@#?(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@!&(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@!|(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@&(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.@|(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);

ALTER OPERATOR __API_CATALOG_SCHEMA__.#=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bsonquery) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.#>(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bsonquery) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.#>= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bsonquery) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.#< (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bsonquery) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);
ALTER OPERATOR __API_CATALOG_SCHEMA__.#<= (__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bsonquery) SET (RESTRICT = __API_SCHEMA_INTERNAL_V2__.bson_dollar_selectivity);