/*-------------------------------------------------------------------------
 * Copyright (c) Microsoft Corporation.  All rights reserved.
 *
 * include/utils/index_utils.h
 *
 * Utilities for build/drop index.
 *
 *-------------------------------------------------------------------------
 */

#ifndef INDEX_UTILS_H
#define INDEX_UTILS_H

void set_indexsafe_procflags(void);
void PopAllActiveSnapshots(void);

#endif
