#include "udfs/rum/composite_path_operator_functions--0.103-0.sql"
#include "schema/composite_path_operator_class--0.103-0.sql"
#include "udfs/projection/bson_expression--0.103-0.sql"

#include "udfs/commands_crud/insert--0.103-0.sql"
#include "udfs/commands_crud/update--0.103-0.sql"