DROP OPERATOR IF EXISTS	__API_CATALOG_SCHEMA__.=?=(__CORE_SCHEMA__.bson, __CORE_SCHEMA__.bson);
CREATE OPERATOR __API_CATALOG_SCHEMA__.=?= (
    LEFTARG = __CORE_SCHEMA__.bson,
    RIGHTARG = __CORE_SCHEMA__.bson,
    PROCEDURE = __API_SCHEMA_INTERNAL_V2__.bson_unique_index_equal,
    COMMUTATOR = OPERATOR(__API_CATALOG_SCHEMA__.=?=)
);