-- show all functions, procedures, aggregates and window aggregates exported in documentdb_api.
\df documentdb_api.*

\df documentdb_api_catalog.*

\df documentdb_api_internal.*

\df documentdb_data.*

-- Access methods + Operator families
\dA *documentdb*

\dAc *documentdb*

\dAf *documentdb*

\dX *documentdb*

-- This is last (Tables/indexes)
\d documentdb_api.*

\d documentdb_api_internal.*

\d documentdb_api_catalog.*

\d documentdb_data.*
