/*-------------------------------------------------------------------------
 * Copyright (c) 2025 documentdb.  All rights reserved.
 *
 * include/bitmap_utils/roaring_bitmap_adapter.h
 *
 * adapter for bitmap functions
 *
 *-------------------------------------------------------------------------
 */

 #ifndef ROARING_BITMAP_ADAPTER_H
 #define ROARING_BITMAP_ADAPTER_H

void RegisterRoaringBitmapHooks(void);

 #endif
